/*
 * Includes
 */
#include "flash.h"

// flash type = GBA Movie Player
unsigned int flashtype = FLASHTYPE_GBAMP;
unsigned int _0 = 0x1E8F;
unsigned int _1 = 0x168F;
unsigned int _5555 = 0x213C2;
unsigned int _2AAA = 0x10C3D;
FlashRead *flashRead;
FlashWrite *flashWrite;

/*
 * Flash address bits of GBA Movie Player
 */
struct GbaMpAddrBits
{
	int bitnum;
	int invert;
};

GbaMpAddrBits addrbits[] =
{
// 0
	{ 10,	1 },		// pin 6
	{ 11,	1 },		// pin 5
	{ 12,	1 },		// pin 4
	{ 8,	1 },		// pin 8
	{ 9,	0 },		// pin 7
	{ 7,	0 },		// pin 18
	{ 6,	0 },		// pin 19
	{ 5,	1 },		// pin 20
// 8
	{ 2,	0 },		// pin 23
	{ 3,	1 },		// pin 22
	{ 4,	1 },		// pin 21
	{ 0,	1 },		// pin 25
	{ 1,	1 },		// pin 24
	{ -1,	0 },		// unused
	{ -1,	0 },		// unused
	{ -1,	0 },		// unused
// 16
	{ 13,	0 },		// pin 3
	{ 14,	0 },		// pin 2
	{ 15,	0 },		// pin 1
	{ 16,	0 },		// pin 48
	{ 17,	0 },		// pin 17
	{ 18,	0 },		// pin 16
};

#define	ADDRBITS	(sizeof(addrbits) / sizeof(addrbits[0]))

/*
 * unlock_mp
 */
unsigned char unlock_mp[] =
{
	0x0158>>1, 0x0064>>1, 0x01C8>>1, 0x00D0>>1,	0x0180>>1, 0x01FC>>1, 0x0150>>1, 0x01C4>>1,
	0x0084>>1, 0x01E0>>1, 0x0154>>1, 0x013C>>1,	0x0024>>1, 0x0178>>1, 0x01B8>>1, 0x0124>>1,
	0x01B4>>1, 0x0154>>1, 0x00E0>>1, 0x01E8>>1,	0x011C>>1, 0x007C>>1, 0x01E0>>1, 0x0000>>1,
	0x001C>>1, 0x0030>>1, 0x0114>>1, 0x01E4>>1,	0x0134>>1, 0x0058>>1, 0x0034>>1, 0x0158>>1,
	0x00B0>>1, 0x01B0>>1, 0x0190>>1, 0x01A4>>1,	0x0114>>1, 0x011C>>1, 0x01EC>>1, 0x00D4>>1,
	0x013C>>1, 0x00F4>>1, 0x002C>>1, 0x0138>>1,	0x004C>>1, 0x0154>>1, 0x00E4>>1, 0x0018>>1,
	0x0120>>1, 0x0144>>1, 0x00F4>>1, 0x0090>>1,	0x01C4>>1, 0x0158>>1, 0x007C>>1, 0x015C>>1,
	0x002C>>1, 0x0040>>1, 0x00EC>>1, 0x01D0>>1,	0x01AC>>1, 0x00DC>>1, 0x01C8>>1, 0x00F8>>1,
	0x017C>>1, 0x0010>>1, 0x0164>>1, 0x0058>>1,	0x0048>>1, 0x0168>>1, 0x0030>>1, 0x00B4>>1,
	0x0064>>1, 0x0000>>1, 0x000C>>1, 0x009C>>1,	0x00C0>>1, 0x01E4>>1, 0x0138>>1, 0x0004>>1,
	0x0188>>1, 0x0054>>1, 0x012C>>1, 0x01BC>>1,	0x01B0>>1, 0x0040>>1, 0x0120>>1, 0x0134>>1,
	0x01A4>>1, 0x01C0>>1, 0x01F4>>1, 0x0110>>1,	0x0080>>1, 0x00F4>>1, 0x00B0>>1, 0x0020>>1,
	0x0044>>1, 0x014C>>1, 0x0130>>1, 0x007C>>1,	0x0024>>1, 0x01D4>>1, 0x005C>>1, 0x0030>>1,
	0x01F8>>1, 0x01F4>>1, 0x00A4>>1, 0x0170>>1,
};





// read cartridge bus
unsigned short CartRead(unsigned int rom_addr)
{
	SetCartAddr(rom_addr);
	return PPReadWord();
}

// write cartridge space
void CartWrite(unsigned int rom_addr, unsigned short data)
{
	SetCartAddr(rom_addr);
	PPWriteWord(data);
}

// read without mirrors
unsigned short GbampRomRead(unsigned int rom_addr)
{
	SetCartAddr(rom_addr >> 13 << 16 | (rom_addr & 0x1FFF));
	return PPReadWord();
}

// write without mirrors
void GbampRomWrite(unsigned int rom_addr, unsigned short data)
{
	SetCartAddr(rom_addr >> 13 << 16 | (rom_addr & 0x1FFF));
	PPWriteWord(data);
}

// read in flash order (slow!)
unsigned short GbampFlashRead(unsigned int flash_addr)
{
	unsigned int cart_addr = FlashToCart(flash_addr);
	#ifdef PRINT_CART_ADDR
		printf("FlashRead(0x%X) -> CartRead(0x%X)\n", flash_addr, cart_addr);
	#endif
	SetCartAddr(cart_addr);
	return PPReadWord();
}

// write in flash order (slow!)
void GbampFlashWrite(unsigned int flash_addr, unsigned short data)
{
	unsigned int cart_addr = FlashToCart(flash_addr);
	#ifdef PRINT_CART_ADDR
		printf("FlashWrite(0x%X, 0x%X) -> CartWrite(0x%X, 0x%X)\n", flash_addr, data, cart_addr, data);
	#endif
	SetCartAddr(cart_addr);
	PPWriteWord(data);
}

/*
 * FlashSetType()
 */
void FlashSetType(unsigned int type, unsigned int rwtype)
{
	flashtype = type;
	switch (type)
	{
		case FLASHTYPE_GBAMP:		// GBA Movie Player
			_0 = 0x1E8F;
			_1 = 0x168F;
			_5555 = 0x213C2;
			_2AAA = 0x10C3D;
			break;

		case FLASHTYPE_SUPERCARD:		// Supercard
			_0 = 0x0;
			_1 = 0x80;
			_5555 = 0x05C9;
			_2AAA = 0x0A36;
			break;
	}

	switch (rwtype)
	{
		case RWTYPE_CART: flashRead = CartRead; flashWrite = CartWrite; break;				
		case RWTYPE_GBAMP_ROM: flashRead = GbampRomRead; flashWrite = GbampRomWrite; break;		
		case RWTYPE_GBAMP_FLASH: flashRead = GbampFlashRead; flashWrite = GbampFlashWrite; break;	
	}
}

/*
 * Supercard_writereg
 */
static void Supercard_writereg(unsigned short x)
{
	CartWrite(0x9FFFFFE>>1, 0xA55A);
	CartWrite(0x9FFFFFE>>1, 0xA55A);
	CartWrite(0x9FFFFFE>>1, x);
	CartWrite(0x9FFFFFE>>1, x);
}

/*
 * FlashUnlock
 */
void FlashUnlock()
{
	if (flashtype == FLASHTYPE_GBAMP)
	{
		for (unsigned int i=0; i<sizeof(unlock_mp) / sizeof(unlock_mp[0]); i++)
		{
			CartRead(unlock_mp[i]);
			CartRead(unlock_mp[i] | 1);
		}
	}
	else if (flashtype == FLASHTYPE_SUPERCARD)
	{
		Supercard_writereg(0x4000 | 0<<4 | 1<<2);
	}
}

/*
 * FlashToCart
 */
unsigned int FlashToCart(unsigned int flash_addr)
{
	unsigned int cart_addr = 0;
	for (unsigned int j=0; j<ADDRBITS; j++)
	{
		if (addrbits[j].bitnum >= 0)
		{
			cart_addr ^= ((flash_addr >> addrbits[j].bitnum & 1) ^ addrbits[j].invert) << j;
		}
	}
	return cart_addr;
}

/*
 * FlashID
 */
unsigned int FlashID()
{
	CartWrite(_5555, 0xAA);
	CartWrite(_2AAA, 0x55);
	CartWrite(_5555, 0x90);
	unsigned short a = CartRead(_0);
	unsigned short b = CartRead(_1);
	CartWrite(_5555, 0xAA);
	CartWrite(_2AAA, 0x55);
	CartWrite(_5555, 0xF0);
	return a | b<<16;
}

/*
 * FlashEraseChip
 */
int FlashEraseChip()
{
	CartWrite(_5555, 0xAA);
	CartWrite(_2AAA, 0x55);
	CartWrite(_5555, 0x80);
	CartWrite(_5555, 0xAA);
	CartWrite(_2AAA, 0x55);
	CartWrite(_5555, 0x10);
	
//	Delay100ms();
	
	unsigned short xo = CartRead(_0);
	while (1)
	{
		unsigned short x = CartRead(_0);
		if ((x & 0x80) && ((x & 0x40) == (xo & 0x40))) break;	// data=1, not toggling
		//printf("%04X\n", x);
		//if (x & 0x80) break;
		xo = x;
	}

	return 0;
}

/*
 * FlashDump
 */
int FlashDump(unsigned short *buffer, unsigned int addr, unsigned int length)
{
	for (unsigned int i=0; i<length; i++)
	{
		buffer[i] = flashRead(addr++);
	}
	return 0;
}

/*
 * FlashVerify
 */
int FlashVerify(unsigned short *buffer, unsigned int addr, unsigned int length)
{
	for (unsigned int i=0; i<length; i++)
	{
		unsigned short data = flashRead(addr);
		if (data != buffer[i]) return -1;
		addr++;
	}
	return 0;
}

/*
 * FlashProgram
 */
int FlashProgram(unsigned short *buffer, unsigned int addr, unsigned int length)
{
	for (unsigned int i=0; i<length; i++)
	{
		CartWrite(_5555, 0xAA);
		CartWrite(_2AAA, 0x55);
		CartWrite(_5555, 0xA0);
		flashWrite(addr, buffer[i]);

		/*unsigned short xo = flashRead(addr);
		while (1)
		{
			unsigned short x = flashRead(addr);
			if (((x & 0x40) == (xo & 0x40))) break;		// not toggling
			xo = x;
		}*/
		//do { RomWrite(addr, buffer[i]); } while (RomRead(addr) != buffer[i]);			// check
		//while ((RomRead(addr) & 0x80) != (buffer[i] & 0x80));			// check
		while (flashRead(addr) != buffer[i]);			// check
		addr++;
	}
	return 0;
}
