/*
	flash.h
*/

/*
 * Includes
 */

#ifdef PRINT_CART_ADDR
	#include <stdio.h>
#endif

/*
 * Defines
 */

#define FLASHID_MANUFACTURER		0x0000FFFF
#define FLASHID_SST					0x000000BF		// SST
#define FLASHID_FAIRCHILD			0x00000004		// Fairchild
#define FLASHID_MX					0x000000C2		// MX

#define FLASHID_SST_512KB			0x278000BF		// SST39xF400A	4 Mib	512 KB
#define FLASHID_SST_1024KB			0x278100BF		// SST39xF800A	8 Mib	1024 KB
#define FLASHID_FAIRCHILD_512KB		0x22B90004		// F 29LV400B	4 Mib	512 KB
#define FLASHID_MX_512KB			0x22B900C2		// MX29LV400B	4 Mib	512 KB

#define FLASHTYPE_GBAMP				0				// FLASHID_SST_512KB, FLASHID_SST_1024KB
#define FLASHTYPE_SUPERCARD			1				// FLASHID_FAIRCHILD_512KB

#define RWTYPE_CART					0	// plain access
#define RWTYPE_GBAMP_ROM			1   // access skipping gaps cause by 3 usused addresslines
#define RWTYPE_GBAMP_FLASH			2   // completely remap

typedef void FlashWrite(unsigned int rom_addr, unsigned short data);
typedef unsigned short FlashRead(unsigned int rom_addr);

/*
 * Imports
 */

void Delay100ms();

#ifdef FLASH_LINKER
	void SetCartAddr(unsigned int addr);
	unsigned short PPReadWord();
	void PPWriteWord(unsigned short data);
#else
	#define SetCartAddr(addr)		volatile unsigned short *p = (volatile unsigned short *)(0x08000000 | (addr) << 1)
	#define PPReadWord()			(*p)
	#define PPWriteWord(data)		*p = data
#endif

/*
 * Variables
 */

extern FlashRead *flashRead;
extern FlashWrite *flashWrite;

/*
 * Prototypes
 */

void FlashSetType(unsigned int type, unsigned int rwtype);
void FlashUnlock();
unsigned int FlashID();
int FlashEraseChip();
int FlashProgram(unsigned short *data, unsigned int addr, unsigned int length);
int FlashVerify(unsigned short *data, unsigned int addr, unsigned int length);
int FlashDump(unsigned short *data, unsigned int addr, unsigned int length);

// for testing purposes
unsigned int FlashToCart(unsigned int flash_addr);
unsigned short CartRead(unsigned int rom_addr);
void CartWrite(unsigned int rom_addr, unsigned short data);
