/*
 * NDS/GBA
 */
#ifdef ARM9        // assume NDS
    #include "nds.h"
    #define ScanKeys        scanKeys
    #define KeysInit        keysInit
    #define KeysHeld        keysHeld
    #define KeysDown        keysDown
    #define KeysUp            keysUp
    #define VBlankIntrWait    swiWaitForVBlank
#else
    #include "gba_base.h"
    #include "gba_input.h"
    #include "gba_interrupt.h"
#endif

/*
 * Includes
 */
    #include <malloc.h>
    #include <nds/memory.h>

#include "flash.h"
#include "console.h"


//#include "loadmp_bin.h"
//#include "loadraw_bin.h"
//#include "repair_bin.h"

//#include "gbamp_bin.h"
#include "supercard_bin.h"


#define FLASH_IDS        3            // max. number of flash IDs

struct Firmware
{
    u32 key;
    char *text;        // 30 chars, parameters: flash id
    const u8 *data;
    u32 size;
    u32 flashtype;
    u32 rwtype;
    u32 flash_ids[FLASH_IDS];
};

Firmware firmwares[] =
{
    { KEY_UP,        "Up    : Supercard\n   \n", supercard_bin, (u32)524288, FLASHTYPE_SUPERCARD, RWTYPE_CART, {FLASHID_FAIRCHILD_512KB, FLASHID_MX_512KB} },
    //{ KEY_START,    "Start : 
    //{ KEY_SELECT,    "Select: 
};

#define FIRMWARES    1 //(sizeof(firmwares) / sizeof(firmwares[0]))

/*
 * Imports
 */
//int xprintf(char *fmt...);
//void gotoxy(int x, int y);

/*
 * FlashApp
 */
int FlashApp(void)
{
    consolePrintf("SC Firmware Header Patcher\n");
    consolePrintf("--------------------------\n");
    consolePrintf("**Warning** this may damage your");
    consolePrintf("SuperCard - it has been tested\n");
    consolePrintf("with SD firmware 1.54 and 1.52\n");
    consolePrintf("But it is not an official tool**\n\n");
    consolePrintf("To patch, hold L+R and press:\n");
    consolePrintf("\n");

    while (1)
    {
        ScanKeys();
        unsigned int held = KeysHeld();
        unsigned int down = KeysDown();

        consolePrintSet(0,9);
        
        for (u32 f=0; f<FIRMWARES; f++)
        {
            Firmware &firmware = firmwares[f];
            
            FlashSetType(firmware.flashtype, firmware.rwtype);
            FlashUnlock();

            u32 id = FlashID();
            consolePrintf(firmware.text, id);

            bool idok = false;
            for (u32 i=0; i<FLASH_IDS; i++) if (firmware.flash_ids[i] && (id == firmware.flash_ids[i])) { idok = true; break; }

            if (idok)
                consolePrintf("   Found!           (%08X)\n\n", id);
/*            else if ((held & (KEY_L | KEY_R)) == (KEY_L | KEY_R))
            {
                consolePrintf("   Override check   (%08X)\n\n", id);
                idok = true;
            }*/

            if (idok)
            {
                if (((held & (KEY_L | KEY_R)) == (KEY_L | KEY_R)) && (down == firmware.key))
                {
                    //copy origional firmware data here
                    consolePrintSet(0,15);
                    consolePrintf("DO NOT SHUT OFF YOUR DS!!!\n");
                    consolePrintf("Reading current firmware...\n");
                    unsigned short int * currFirm;
                    currFirm = (unsigned short int *)malloc(524288);
                    FlashDump(currFirm, 0, 524288);

                    // patch the header with PASS and proper GBA CRC
                    consolePrintf("Patching header...\n");
                    currFirm[86]= 0x4150; // PA
                    currFirm[87]= 0x5353; // SS
                    currFirm[94]= 0xF800; // crc

                    // erase the chip - must be done before writing it, hence the copy above
                    consolePrintSet(0,19);
                    consolePrintf("Erasing...\n");
                    FlashEraseChip();

                    int verify = -2;
                    while (verify < 0)
                    {
                        if (down == firmware.key)
                        {
                            //if (verify == -2) 
                            consolePrintf("Programming %d bytes...\n", firmware.size);
                            FlashProgram((u16 *)currFirm, 0, 524288/2);
                            //if (verify == -2) 
                            consolePrintf("Verifying...\n");
                            verify = FlashVerify((u16 *)currFirm, 0, 524288/2);    // 0 or -1
                        }
                        if (verify < 0) for (int i=0; i<10; i++) VBlankIntrWait();        // wait before retry
                    }
                    consolePrintf("Done!!! OK to shut down NOW!!");
                    return 0;
                }
            }
            else
            {
                consolePrintf("   Not found!       (%08X)\n\n", id);
            }
        }    // firmwares
    }
    return 1;
}
